local _, BigBrother = ...
local SpellData = BigBrother.SpellData

local module = BigBrother:NewModule("BuffCheck")
local L = LibStub("AceLocale-3.0"):GetLocale("BigBrother")

local _G = _G

local CLASS_COLORS = {}

local DEFAULT_ROWS = 10
local DEFAULT_ROWS_MIN = 5
local DEFAULT_ROWS_MAX = 25

local BUFF_SIZE = 16
local BUFF_SPACING = 18
local TOTAL_BUFFS = 8
local BAR_HEIGHT = 18
local BAR_WIDTH = 290

local BuffWindow, BuffTable
local NumPlayersShown, NumRowsCreated
local SelectedPanel = 1
local PlayerList = {}
local DirtyBuffs

local CreateBuffRow, CreateBuffWindow, ResizeBuffWindow, UpdateBuffWindow, UpdateBuffWindowBuffs

local MaxFood = {}
for _, v in next, SpellData.foodsMax do
	MaxFood[v] = true
end

local DefaultRaidBuffs = {
	1126, --Mark of the Wild
	21562, --PW:Fortitude
	6673, --Battle Shout
	30809, -- Unleashed Rage
	1459, -- Arcane Brilliance
	24907, -- Moonkin Aura
	1459, -- Arcane Brilliance
	19740, --Blessing of Might
}

do --BuffTable
	local function spellData(spellId)
		local name = GetSpellInfo(spellId) or ""
		return name
	end

	--Load up local tables from master spell ID tables
	local function createBuffGroup(tbl)
		local t = {}
		for i,v in ipairs(tbl) do
			t[i] = spellData(v)
		end
		return t
	end

	local function buffSort(a, b)
		if a.buffs < b.buffs then
			return true
		elseif a.buffs > b.buffs then
			return false
		elseif a.name < b.name then
			return true
		end
		return false
	end

	local function classSort(a, b)
		if a.class < b.class then
			return true
		elseif a.class > b.class then
			return false
		elseif a.name < b.name then
			return true
		end
		return false
	end

	BuffTable = {
		{
			name = L["Raid Buffs"],
			sortFunc = buffSort,
			buffs = {
				{ --Stats
					spellData(1126), --Mark of the Wild
					spellData(20217), --Blessing of Kings
					spellData(90363), --Embrace of the Shale Spider (pet)
					spellData(117666), --Legacy of the Emperor
				},
				{ --Stamina
					spellData(469), --Commanding Shout
					spellData(6307), --Blood Pact (pet)
					spellData(21562), --PW:Fortitude
					spellData(90364), --Qiraji Fortitude (pet)
				},
				{ --Attack Power
					spellData(6673), --Battle Shout
					spellData(19506), --Trueshot Aura
					spellData(57330), --Horn of Winter
				},
				{ --Attack Speed
					spellData(30809), -- Unleashed Rage
					spellData(55610), -- Unholy Aura
					spellData(113742), -- Swiftblade's Cunning
					spellData(128432), -- Cackling Howl (pet)
					spellData(128433), -- Serpent's Swiftness (pet)
				},
				{ --Spell Power
					spellData(1459), -- Arcane Brilliance
					spellData(61316), -- Dalaran Brilliance
					spellData(77747), -- Burning Wrath
					spellData(109773), -- Dark Intent
					spellData(126309), -- Still Water (pet)
				},
				{ --Spell Haste
					spellData(24907), -- Moonkin Aura
					spellData(49868), -- Mind Quickening
					spellData(51470), -- Elemental Oath
					spellData(135678), -- Energizing Spores (pet)
				},
				{ --Critical Strike
					spellData(1459), -- Arcane Brilliance
					spellData(61316), -- Dalaran Brilliance
					spellData(17007), -- Leader of the Pack
					spellData(116781), -- Legacy of the White Tiger
					spellData(24604), -- Furious Howl (pet)
					spellData(90309),-- Terrifying Roar (pet)
					spellData(97229), -- Bellowing Roar (pet)
					spellData(126309), -- Still Water (pet)
					spellData(126373), -- Fearless Roar (pet)
				},
				{ --Mastery
					spellData(19740), --Blessing of Might
					spellData(93435), --Roar of Courage (pet)
					spellData(116956), --Grace of Air
					spellData(128997), --Spirit Beast Blessing (pet)
				},
			},
		}, --[1]
		{
			name = L["Consumables"],
			sortFunc = classSort,
			buffs = {
				{}, {}, {}, {}, --wish there was something else related and useful to fill these columns
				createBuffGroup(SpellData.flasks),
				createBuffGroup(SpellData.elixirsBattle),
				createBuffGroup(SpellData.elixirsGuardian),
				createBuffGroup(SpellData.foods),
			}
		}, --[2]
	}
end


local function GetHeightFromRows(rows)
	return 4 + BUFF_SPACING + 4 + ((BAR_HEIGHT + 2) * rows) + 4 + 0.5
end

local function GetRowsFromHeight(height)
	return math.floor( (height - (BUFF_SPACING + 4) - 8) / (BUFF_SPACING + 2) )
end

local function SetDirty()
	DirtyBuffs = true
end

local function BuffWindow_OnEnter(frame)
	GameTooltip:SetOwner(frame, "ANCHOR_TOPLEFT")
	if frame.name then
		GameTooltip:SetUnitBuff(frame:GetParent().unit, frame.name)
	elseif frame.tooltip then
		GameTooltip:AddLine(frame.tooltip)
	end
	GameTooltip:Show()
end

local function BuffWindow_OnLeave()
	GameTooltip:Hide()
end

local BuffWindow_OnUpdate
do --BuffWindow_OnUpdate
	local total = 0
	function BuffWindow_OnUpdate(self, elapsed)
		total = total + elapsed
		if total > 0.5 and DirtyBuffs then
			UpdateBuffWindowBuffs()
			total = 0
			DirtyBuffs = nil
		end
	end
end

function CreateBuffRow(parent, xOffset, yOffset)
	local row = CreateFrame("Frame", nil, parent)
	row:SetPoint("TOPLEFT", xOffset, yOffset)
	row:SetHeight(BAR_HEIGHT)
	row:SetWidth(BAR_WIDTH - 16) --account for the scroll bar

	local bg = row:CreateTexture(nil, "BACKGROUND")
	bg:SetAllPoints()
	bg:SetTexture("Interface\\Buttons\\WHITE8X8.blp")
	bg:SetGradientAlpha("HORIZONTAL", 0.5, 0.0, 0.0, 0.8, 0.5, 0.0, 0.0, 0.0)
	row.background = bg

	row.buff = {}
	local last
	for i = TOTAL_BUFFS, 1, -1 do
		local buff = CreateFrame("Frame", nil, row)
		if last then
			buff:SetPoint("RIGHT", last, "LEFT", -2, 0)
		else
			buff:SetPoint("RIGHT", -2, 0)
		end
		buff:SetHeight(BUFF_SIZE)
		buff:SetWidth(BUFF_SIZE)

		local icon = buff:CreateTexture(nil, "OVERLAY")
		icon:SetAllPoints()
		icon:SetTexture("Interface\\Buttons\\UI-CheckBox-Check.blp")
		icon:SetTexCoord(.07, .93, .07, .93)
		buff.icon = icon

		buff:SetScript("OnEnter", BuffWindow_OnEnter)
		buff:SetScript("OnLeave", BuffWindow_OnLeave)
		buff:EnableMouse()

		last = buff
		row.buff[i] = buff
	end

	local name = row:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	name:SetPoint("LEFT", 2, 0)
	name:SetPoint("RIGHT", last, "LEFT", -2, 0)
	name:SetJustifyH("LEFT")
	name:SetTextColor(1.0, 1.0, 1.0)
	name:SetText(_G.UNKNOWN)
	row.name = name

	return row
end

function CreateBuffWindow()
	local frame = CreateFrame("Frame", "BigBrotherBuffsFrame", UIParent)
	frame:Hide()

	--set on top of most ui elements
	frame:SetFrameStrata("HIGH")
	frame:SetFrameLevel(10)

	frame:ClearAllPoints()
	if module.db.profile.x and module.db.profile.y then
		frame:SetPoint("TOPLEFT", UIParent, "BOTTOMLEFT", module.db.profile.x, module.db.profile.y)
	else
		frame:SetPoint("CENTER")
	end
	local height = module.db.profile.height or GetHeightFromRows(DEFAULT_ROWS)

	frame:SetBackdrop({
		bgFile = "Interface/Tooltips/UI-Tooltip-Background",
		edgeFile = "Interface/Tooltips/UI-Tooltip-Border",
		tile = true, tileSize = 16, edgeSize = 16,
		insets = { left = 4, right = 4, top = 4, bottom = 4 }
	})
	frame:SetBackdropColor(0, 0, 0, 0.5)

	frame:EnableMouse()
	frame:SetMovable(true)
	frame:SetClampedToScreen(true)
	frame:SetResizable(true)
	frame:SetMinResize(BAR_WIDTH + 24, GetHeightFromRows(DEFAULT_ROWS_MIN))
	frame:SetMaxResize(BAR_WIDTH + 24, GetHeightFromRows(DEFAULT_ROWS_MAX))
	frame:SetWidth(BAR_WIDTH + 24)
	frame:SetHeight(height)

	frame:SetScript("OnMouseDown", function(self, button)
		if button == "LeftButton" then
			self:StartMoving()
			self.isMoving = true
		end
	end)
	frame:SetScript("OnMouseUp", function(self)
		if self.isMoving then
			self:StopMovingOrSizing()
			self.isMoving = false
			module.db.profile.x = self:GetLeft()
			module.db.profile.y = self:GetTop()
		end
	end)
	frame:SetScript("OnHide", function(self)
		if self.isMoving then
			self:StopMovingOrSizing()
			self.isMoving = false
		end
	end)
	frame:SetScript("OnShow", UpdateBuffWindowBuffs)
	frame:SetScript("OnUpdate", BuffWindow_OnUpdate)
	frame:SetScript("OnSizeChanged", function(self)
		if self.isResizing then
			ResizeBuffWindow()
		end
	end)

	local title = frame:CreateFontString(nil, "OVERLAY", "GameFontNormal")
	title:SetPoint("TOP", frame, "TOP", 0, -8)
	title:SetTextColor(1.0, 1.0, 1.0)
	frame.title = title

	local leftButton = CreateFrame("Button", nil, frame)
	leftButton:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Up.blp")
	leftButton:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-PrevPage-Down.blp")
	leftButton:SetWidth(16)
	leftButton:SetHeight(18)
	leftButton:SetPoint("TOPLEFT", frame, "TOPLEFT", 64, -5)
	leftButton:SetScript("OnClick", function()
		SelectedPanel = SelectedPanel - 1
		if SelectedPanel == 0 then
			SelectedPanel = #BuffTable
		end
		UpdateBuffWindowBuffs()
	end)

	local rightButton = CreateFrame("Button", nil, frame)
	rightButton:SetNormalTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Up.blp")
	rightButton:SetPushedTexture("Interface\\Buttons\\UI-SpellbookIcon-NextPage-Down.blp")
	rightButton:SetWidth(16)
	rightButton:SetHeight(18)
	rightButton:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -64, -5)
	rightButton:SetScript("OnClick", function()
		SelectedPanel = SelectedPanel + 1
		if SelectedPanel > #BuffTable then
			SelectedPanel = 1
		end
		UpdateBuffWindowBuffs()
	end)

	local closeButton = CreateFrame("Button", nil, frame)
	closeButton:SetNormalTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Up.blp")
	closeButton:SetPushedTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Down.blp")
	closeButton:SetHighlightTexture("Interface\\Buttons\\UI-Panel-MinimizeButton-Highlight.blp")
	closeButton:SetWidth(22)
	closeButton:SetHeight(22)
	closeButton:SetPoint("TOPRIGHT", frame, "TOPRIGHT", -4, -4)
	closeButton:SetScript("OnClick", function(self) self:GetParent():Hide() end)

	local readyButton = CreateFrame("Button", nil, frame)
	readyButton:SetNormalTexture("Interface\\RAIDFRAME\\ReadyCheck-Waiting")
	readyButton:SetWidth(12)
	readyButton:SetHeight(12)
	readyButton:SetPoint("TOPLEFT", frame, "TOPLEFT", 8, -8)
	readyButton:SetScript("OnClick", function() DoReadyCheck() end)
	readyButton:SetScript("OnEnter", BuffWindow_OnEnter)
	readyButton:SetScript("OnLeave", BuffWindow_OnLeave)
	readyButton.tooltip = _G.READY_CHECK

	local checkButton = CreateFrame("Button", nil, frame)
	checkButton:SetNormalTexture("Interface\\RAIDFRAME\\ReadyCheck-Ready")
	checkButton:SetWidth(12)
	checkButton:SetHeight(12)
	checkButton:SetPoint("TOPLEFT", frame, "TOPLEFT", 22, -8)
	checkButton:SetScript("OnClick", function() BigBrother:ConsumableCheck() end)
	checkButton:SetScript("OnEnter", BuffWindow_OnEnter)
	checkButton:SetScript("OnLeave", BuffWindow_OnLeave)
	checkButton.tooltip = L["Quick Check"]

	local rows = {}
	for i = 1, GetRowsFromHeight(height) do
		rows[i] = CreateBuffRow(frame, 8, -4 - i * (BUFF_SPACING + 2))
	end
	frame.rows = rows
	NumRowsCreated = #rows
	NumPlayersShown = #rows

	local scrollBar = CreateFrame("ScrollFrame", "BigBrotherBuffsFrameScrollFrame", frame, "FauxScrollFrameTemplate")
	scrollBar:SetScript("OnVerticalScroll", function(self, offset)
		FauxScrollFrame_OnVerticalScroll(self, offset, BAR_HEIGHT + 2, UpdateBuffWindow)
	end)
	scrollBar:SetPoint("TOPLEFT", frame, "TOPLEFT", 8, -24)
	scrollBar:SetPoint("BOTTOMRIGHT", frame, "BOTTOMRIGHT", -8 - 24, 8)
	frame.scrollBar = scrollBar

	local dragHandle = CreateFrame("Frame", nil, frame)
	dragHandle:Show()
	dragHandle:SetFrameLevel(frame:GetFrameLevel() + 10) -- place this above everything
	dragHandle:SetWidth(BAR_WIDTH + 24)
	dragHandle:SetHeight(8)
	dragHandle:SetPoint("BOTTOM", frame, "BOTTOM", 0, 0)
	dragHandle:EnableMouse(true)
	dragHandle:SetBackdrop({
		bgFile = "Interface\\Tooltips\\UI-Tooltip-Background",
		tile = true, tileSize = 16, edgeSize = 0,
		insets = { left = 6, right = 6, top = 3, bottom = 3 }
	})
	dragHandle:SetBackdropColor(1, 1, 1, 0)
	dragHandle:SetScript("OnEnter", function(self)
		self:SetBackdropColor(1, 1, 1, 0.8)
	end)
	dragHandle:SetScript("OnLeave", function(self)
		self:SetBackdropColor(1, 1, 1, 0)
	end)
	dragHandle:SetScript("OnMouseDown", function(self, button)
		if button == "LeftButton" then
			self:GetParent().isResizing = true
			self:GetParent():StartSizing("BOTTOMRIGHT")
		end
	end)
	dragHandle:SetScript("OnMouseUp", function(self, button)
		if button == "LeftButton" then
			self:GetParent():StopMovingOrSizing()
			self:GetParent().isResizing = false
			self:GetParent():SetHeight(GetHeightFromRows(NumPlayersShown))
		elseif button == "RightButton" then
			self:GetParent():SetHeight(GetHeightFromRows(DEFAULT_ROWS))
		end
		ResizeBuffWindow()
	end)

	return frame
end

--update the player buff table
do --UpdateBuffWindowBuffs
	local partyUnits = {"player"}
	for i = 1, MAX_PARTY_MEMBERS do
		partyUnits[#partyUnits + 1] = ("party%d"):format(i)
	end
	local raidUnits = {}
	for i = 1, MAX_RAID_MEMBERS do
		raidUnits[#raidUnits + 1] = ("raid%d"):format(i)
	end

	function UpdateBuffWindowBuffs()
		local groups = BigBrother:GetCheckedGroups()
		local currentZone = GetRealZoneText()

		local inRaid = IsInRaid(LE_PARTY_CATEGORY_HOME)
		local units = inRaid and raidUnits or partyUnits

		local index = 1
		for i, unit in ipairs(units) do
			local name, subgroup, class, zone, online, dead
			if inRaid then
				name, _, subgroup, _, _, class, zone, online, dead = GetRaidRosterInfo(i)
			else
				name = UnitName(unit)
				subgroup = 1
				class = select(2, UnitClass(unit)) or _G.UNKNOWN --cross-realm players take a sec to load, don't break sorting prz
				online = UnitIsConnected(unit)
				dead = UnitIsDeadOrGhost(unit)
			end
			if not name then break end
			name = gsub(name, "%-.*", "*")

			if groups[subgroup] then
				local player = PlayerList[index]
				if not player then
					player = { buff = {} }
				end
				PlayerList[index] = player

				player.name = name
				player.class = class
				player.status = not online or dead or (zone and zone ~= currentZone)
				player.unit = unit
				player.buffs = 0

				for column, buffGroups in ipairs(BuffTable[SelectedPanel].buffs) do
					player.buff[column] = nil
					for _, buff in ipairs(buffGroups) do
						local spellId = select(11, UnitBuff(player.unit, buff))
						if spellId then
							player.buff[column] = spellId
							player.buffs = player.buffs + 1
							break
						end
					end
				end

				index = index + 1
			end
		end

		-- clear the rest of the table so we dont get nil holes that lead to ill-defined behavior in sort
		while PlayerList[index] do
			PlayerList[index] = nil
			index = index + 1
		end
		sort(PlayerList, BuffTable[SelectedPanel].sortFunc)

		UpdateBuffWindow()
	end
end

--update the display
function UpdateBuffWindow()
	BuffWindow.title:SetText(BuffTable[SelectedPanel].name)

	FauxScrollFrame_Update(BuffWindow.scrollBar, #PlayerList, NumPlayersShown, 20)
	local offset = FauxScrollFrame_GetOffset(BuffWindow.scrollBar)

	for i = 1, NumPlayersShown do
		local row = BuffWindow.rows[i]
		if PlayerList[i + offset] then
			local player = PlayerList[i + offset]
			row.unit = player.unit
			local color = not player.status and _G.HIGHLIGHT_FONT_COLOR or _G.RED_FONT_COLOR
			local bgColor = CLASS_COLORS[player.class] or _G.GRAY_FONT_COLOR
			row.name:SetText(player.name)
			row.name:SetTextColor(color.r, color.g, color.b)
			row.background:SetGradientAlpha(
				"HORIZONTAL",
				bgColor.r/1.5, bgColor.g/1.5, bgColor.b/1.5, 0.8,
				bgColor.r/1.5, bgColor.g/2, bgColor.b/1.5, 0
			)

			for column = 1, TOTAL_BUFFS do
				local buff = row.buff[column]
				local spellId = player.buff[column]
				if spellId then
					local name, _, texture = GetSpellInfo(spellId)
					buff.name = name
					buff.tooltip = nil
					buff.icon:SetTexture(texture)
					if texture == "Interface\\Icons\\Spell_Misc_Food" and not MaxFood[spellId] then --texture == "Interface\\Icons\\Spell_Misc_Food"
						buff.icon:SetVertexColor(1, .5, .5) --red tint
					else
						buff.icon:SetVertexColor(1, 1, 1)
					end
					buff.icon:SetDesaturated(nil)
					buff.icon:SetAlpha(1)
					buff:Show()
				else
					if SelectedPanel == 1 then --dimmed icons for missing raid buffs
						buff.name = nil
						buff.tooltip = _G.GRAY_FONT_COLOR_CODE .. _G["RAID_BUFF_"..column]
						local _, _, texture = GetSpellInfo(DefaultRaidBuffs[column])
						buff.icon:SetTexture(texture)
						buff.icon:SetVertexColor(1, 1, 1)
						buff.icon:SetDesaturated(true)
						buff.icon:SetAlpha(0.2)
						buff:Show()
					else
						buff:Hide()
					end
				end
			end
			row:Show()
		else
			row:Hide()
		end
	end
end

function ResizeBuffWindow()
	--clamp the list size
	if module.db.profile.limitResize then
		local minSize, maxSize = GetHeightFromRows(DEFAULT_ROWS_MIN), GetHeightFromRows(#PlayerList)
		if BuffWindow:GetHeight() > maxSize then
			BuffWindow:SetHeight(maxSize < minSize and minSize or maxSize)
		end
	end

	local numVisibleRows = GetRowsFromHeight(BuffWindow:GetHeight())

	if numVisibleRows > NumRowsCreated then
		for i = (1 +  NumRowsCreated), numVisibleRows do
			BuffWindow.rows[i] = CreateBuffRow(BuffWindow, 8, -4 - i * (BUFF_SPACING + 2))
			BuffWindow.rows[i]:Hide()
		end
		NumRowsCreated = numVisibleRows
	end

	if numVisibleRows < NumPlayersShown then
		for i = (1 + numVisibleRows), NumPlayersShown do
			BuffWindow.rows[i]:Hide()
		end
	end
	NumPlayersShown = numVisibleRows

	UpdateBuffWindow()

	module.db.profile.height = BuffWindow:GetHeight()
end


function module:ToggleWindow()
	if BuffWindow and BuffWindow:IsShown() then
		BuffWindow:Hide()
	elseif BuffWindow then
		BuffWindow:Show()
	else
		BuffWindow = CreateBuffWindow()
		BuffWindow:Show()
	end
end

function module:OnInitialize()
	self.db = BigBrother.db:RegisterNamespace("BuffCheck", {
		profile = {
			limitResize = false,
			output = "GROUP",
		}
	})
end

function module:OnEnable()
	self:RegisterEvent("UNIT_AURA", SetDirty)
	self:RegisterEvent("GROUP_ROSTER_UPDATE", SetDirty)

	if _G.CUSTOM_CLASS_COLORS then
		local function updateClassColors()
			for k,v in next, _G.CUSTOM_CLASS_COLORS do
				CLASS_COLORS[k] = v
			end
		end
		_G.CUSTOM_CLASS_COLORS:RegisterCallback(updateClassColors)
		updateClassColors()
	else
		CLASS_COLORS = _G.RAID_CLASS_COLORS
	end
end

function module:AddOptions(options)
	options.args.buffcheck = {
		name = L["Buff Check"],
		desc = L["Pops up a window to check raid buffs (drag the bottom to resize)."],
		type = "execute",
		func = self.ToggleWindow,
		order = 8,
		width = "full",
	}

	options.args.buffwindow = {
		name = L["Buff Check"],
		--desc = L["Buff Check"],
		type = "group",
		get = function(info) return self.db.profile[info[#info]] end,
		set = function(info, value)
			self.db.profile[info[#info]] = value
			SetDirty()
		end,
		order = 30,
		args = {
			limitResize = {
				name = L["Limit size"],
				desc = L["Limits the number of rows shown to the number of people in your group (minimum: %d, maximum: %d)."]:format(DEFAULT_ROWS_MIN, DEFAULT_ROWS_MAX),
				type = "toggle",
				order = 10,
			},
		}
	}
end
